#ifndef _ADIFF
#define _ADIFF
/*
  clear; g++ -D TEST_ADIFF -g adiff.cc -o adiff
  Algorithmic differentation -- forward mode.
 */

#include<iostream>
#include<map>
#include<vector>
#include<math.h>

/*
 */
class var_t;
typedef const var_t& arg_t; 
typedef map<const var_t*,double> deps_t;

class var_t
{
public:
  double val;
  // List of variables on which *this depends.
  // Together with the corresponding partial derivatives.
  // (d *this / d xi)_{i\in I_y1}
  deps_t deps;
public:
  // Copy-constructor:
  var_t( const var_t& x)
  {
    deps = x.deps;
    val = x.val;
  }
  
  var_t()
  {}

  // constant initialized variables:
  double operator = (double _val)
  {
    deps.clear();
    deps[this] = 1.0;
    val = _val;
  }

  /*
    Two usefull constructors for functions in dependents of one or two
    arguments:
  */
  var_t(double z, double dzdx,  const var_t& x)
  {
    deps.clear();
    val = z;
    J(dzdx,x);
  }
  
  var_t(double z_xy,
	double dzdx,  arg_t x,
	double dzdy,  arg_t y)
  {
    deps.clear();
    val = z_xy;
    J(dzdx, x);
    J(dzdy, y);
  }

  /*
    Check whether *this depends on some const initialized variable v.
  */
  bool depents_on(arg_t v)
  {return (deps.find(&v) != (deps_t::const_iterator)deps.end());}

  double by_d(arg_t v) 
  {
    deps_t::const_iterator iter;
    iter = deps.find(&v);
    if( iter == deps.end() )
      return 0.0;
    else
      return iter->second;
  }

  double operator()() const
  {return val;}

  var_t& J(double par_der, arg_t v)
  {
    deps_t::const_iterator iter_v;
    deps_t::iterator iter;
    for( iter_v = v.deps.begin(); iter_v != v.deps.end(); iter_v++ )
      {
	iter = deps.find( iter_v->first );
	if( iter != deps.end() )
	  (iter->second) += par_der * (iter_v->second);
	else
	  deps[ iter_v->first ] = par_der * (iter_v->second);
      }

    return *this;
  }

  // initialization from istream...
  friend istream& operator >> (istream&, var_t&);
};

// initialization from istream...
istream& operator >> (istream& is, var_t& x)
{
  double val;
  is >> val;
  x = val;
  return is;
}

/***********************************************************************
  Some numerical operations:
***********************************************************************/

inline var_t operator + (arg_t x,  arg_t y)
{
  return var_t( x()+y(), 1.0, x, 1.0, y );
}

inline var_t operator + (arg_t x, double y)
{
  return var_t(x()+y, 1.0, x);
}

inline var_t operator + (double x, arg_t y)
{
  return y+x;
}

inline var_t operator * ( arg_t x,  arg_t y)
{
  return var_t( x()*y(), y(), x, x(), y );
}

inline var_t operator * ( arg_t x,  double y)
{
  return var_t( x()*y, y, x);
}

inline var_t operator * (double x, arg_t y)
{ return y*x; }

inline var_t operator / ( arg_t x,  arg_t y)
{
  double z = 1.0/y();
  return var_t( x()*z, z, x, -x()*z*z, y );
}

inline var_t operator / ( double x,  arg_t y)
{
  double z = 1.0/y();
  return var_t(x*z,-z*z,y);
}

inline var_t operator / (arg_t x, double y )
{ return x*(1.0/y); }

inline var_t operator - (arg_t x)
{
  return var_t(-x(),-1.0,x);
}

inline var_t operator - (arg_t x, arg_t y)
{
  return var_t(x()-y(),1.0,x,-1.0,y);
}

inline var_t operator - (double x, arg_t y)
{
  return var_t(x-y(),-1.0,y);
}

inline var_t operator - (arg_t x, double y)
{
  return var_t(x()-y,1.0,x);
}

// unary operations / functions:

inline var_t exp(arg_t x)
{
  double val=exp(x());
  return var_t(val,val,x);
}

inline var_t log(arg_t x)
{
  return var_t(log(x()),1.0/x(),x);
}

inline var_t sin(arg_t x)
{
  return var_t(sin(x()),cos(x()),x);
}

var_t cos(arg_t x)
{
  return var_t(cos(x()),-sin(x()),x);
}

var_t sqrt(arg_t x)
{
  double z = sqrt(x());
  return var_t(z,0.5/z,x);
}
/**********************************************************************
  Test
**********************************************************************/
#ifdef TEST_ADIFF

#include<fstream>

main()
{
  var_t x,y,z,z2,z3;

  std::cout << " x= "; std::cin >> x;
  std::cout << " y= "; std::cin >> y;
  
  z = x * exp(2.0*y);

  z2 = z/x;

  z3 = log(z2);

  std::cout << " z = x * exp(y) = " << z() << std::endl;

  std::cout << " dz/dx = " << z.by_d(x) << std::endl;

  std::cout << " dz/dy = " << z.by_d(y) << std::endl;

  std::cout << " z2 = " << z2() << "\n z3 = " << z3() << std::endl;
  
  std::cout << " dz2/dx = " << z2.by_d(x) << "\n dz2/dy = " << z2.by_d(y) << std::endl;

  std::cout << " dz3/dx = " << z3.by_d(x) << "\n dz3/dy = " << z3.by_d(y) << std::endl;

  std::ofstream os("test.dat");
  int N=101;
  int i;
  double x_min=-1.0;
  double x_max=1.0;
  
  double dx=(x_max-x_min)/(N-1);

  for(i=0, x=x_min; i<N; i++, x = x()+dx)
    {
      y=sqrt(x*x);
      os << x() << ", " << y() << ", " << y.by_d(x) << std::endl;
    }
  os.close();

  std::cout << "\n Some really strange abuse of the adiff package:";
  std::cout << "\n C++ - Code: { var_t x;  x = 1.0; x = exp(x); }";
    x = 1.0; x = exp(x);
  std::cout << "\n x= " << x() << "\n x_by_dx = " << x.by_d(x) << std::endl;

}

#endif // DEF TEST_ADIFF
#endif // NDEF _ADIFF
