#include"tobiMath.hh"
// for debugging:
// int my_double_count = 0;

double pow(double x,int y) {
  double retval=1.0;
  int i,n=abs(y);
  for(i=0;i<n;i++)
    retval*=x;
  return (y<0? 1.0/retval : retval);
}

double** new_matrix(int n, int m)
{
  int i,j;
  double** ptr;
  ptr = new double* [n];
  if(ptr==NULL) return NULL;

  for(i=0;i<n;i++)
    {
      ptr[i] = new double [m];
      if(ptr[i]==NULL)
	{
	  for(j=0;j<i;j++)
	    delete [] ptr[i];
	  delete [] ptr;
	  return NULL;
	}
    }
  return ptr;
}

void delete_matrix(int n, double** A)
{
  int i;
  for(i=0;i<n;i++)
    delete [] A[i];
  delete [] A;
}

void copy_matrix(int n, int m, double** A, double** B)
{
  int i,j;
  for(i=0;i<n;i++)
    for(j=0;j<m;j++)
      A[i][j]=B[i][j];
}

double norm_inf(int dim, double* v)
{
  int k,index_max = 0;
  for(k=1;k<dim;k++)
    if( fabs(v[index_max])<fabs(v[k]) ) index_max = k;

  return fabs(v[index_max]);
}

double* sort_abs(int n, double* x)
{
  int i,j,max;
  double tmp;

  for(i=0;i<n-1;i++)
    {
      max=i;
      for(j=i+1;j<n;j++)
	if(fabs(x[max])<fabs(x[j])) max=j;
      if(max != i)
	{ tmp = x[max]; x[max]=x[i]; x[i]=tmp; }
    }
  return x;
}

double norm2(int dim, double* v)
{
  int k; double retval;
  for(k=0;k<dim;k++)
    retval += v[k]*v[k];
  return sqrt(retval);
}

bool lin_solve_gauss(int dim, double** A, double* b)
{
  int k,l,m,index_max;
  double tmp;

    // Solve the linear system: b = A * x
    // afterward x is stored into b
    // forward sweep:
    for(k=0; k<dim-1; k++)
      {
	// pivotization:
	int index_max=k;
	for(l=k+1; l<dim; l++)
	  if( fabs(A[l][k]) > fabs(A[index_max][k]) )
	    index_max = l;

	if(fabs(A[index_max][k])==0.0)
	  return false;

	if( index_max != k )
	  {
	    for( l=k; l<dim; l++)
	    {
	      tmp = A[k][l];
	      A[k][l]=A[index_max][l];
	      A[index_max][l]=tmp;
	    }
	    tmp = b[k];
	    b[k]=b[index_max];
	    b[index_max]=tmp;
	  }
	
	for(l=k+1;l<dim; l++)
	  {
	    tmp = A[l][k]/A[k][k];
	    for(m=k+1;m<dim;m++)
	      A[l][m] -= tmp*A[k][m];
	    b[l] -= tmp*b[k];
	  }
	
      }
    
    for(k=dim-1;k>=0;k--){//backward sweep
      for(l=k+1; l<dim; l++)
	b[k] -= A[k][l]*b[l];

      b[k] /= A[k][k];
    }

    return true;
}

bool lin_solve_cholesky(int n, double** A, double* b)
{
  /* 
     1st:
     Decomposition of A into L * L', where
     L is a left lower triangular matrix and L' is the transposed of L.
     L is stored into the left lower triangle of A.
     2nd:
     Solution of the system L * L' x = b.
     The result is stored into b.
  */
  int i,j,k;
  double tmp1,tmp2;
  
  for(i=0;i<n;i++) {
    tmp1=A[i][i];
    for(j=0;j<i;j++) {
      tmp2=A[i][j];
      for(k=0;k<j;k++)
	tmp2-=A[i][k]*A[j][k];
      A[i][j]=tmp2/A[j][j];
      tmp1 -= A[i][j]*A[i][j];
    }
    A[i][i]=sqrt(tmp1);
    if(tmp1 == 0.0) return false;
  }

  // Solution of L*x=b, store solution x into b:
  for(i=0;i<n;i++)
    {
      tmp1=0.0;
      for(j=0;j<i;j++)
	tmp1 +=A[i][j]*b[j];
      b[i]=(b[i]-tmp1)/A[i][i];
    }
  // Solution of L'*x=b, store solution x into b. That's it.
  for(i=n-1;i>=0;i--)
    {
      tmp1=0.0;
      for(j=n-1;j>i;j--)
	tmp1 +=A[j][i]*b[j];
      b[i]=(b[i]-tmp1)/A[i][i];
    }
  return true;
}

decomp_householder_pivot_t decomp_householder_pivot = sometimes;
double decomp_householder_eps=1.0e-10;
int decomp_householder(int n, int m, double** A, double* d, int* p )
  /*
    n: number of rows of A,
    m: input value: number of columns of A, m less equal n
       output value: rank of A
    A: input value: system matrix
       output value: Householder-QR-decomposition of A,
       A[k][k]...A[n][k]: k-th reflection vector
       A[0][k]...A[k-1][k]: k-th superdiagonal elements of R
       d[0]...d[m]: diagonal elements of R
    d: diagonal of R
    p: permutation vector of the columns of A.

    return value: rank(A)
    If n=m and the return value differs from n then A is singular.
    The input n<m is not admissible. Therefore the procedure exists
    with return value -1.
   */
{
  int i,j,k;
  int i_max;
  double tmp,tmp1,tmp2;

  if(n<m) return -1;

  for(i=0;i<m;i++) p[i]=i;

  for(i=0;i<m;i++) { // main loop
    switch(decomp_householder_pivot){
    case sometimes:
      for(tmp1=0.0,j=i+1;j<n;j++)
	tmp1 += A[j][i]*A[j][i];

      tmp = sqrt(A[i][i]*A[i][i]+tmp1);
      if(tmp < decomp_householder_eps)
	{
	  // column exchange, if || A[.][j] || > eps for j in i+1 ... m
	  if(i==m-1) return i;

	  for(j=i+1;j<m;j++)
	    {
	      for(tmp1=0.0,k=i+1;k<n;k++)
		tmp1 += A[k][j]*A[k][j];
	      tmp = sqrt(A[i][j]*A[i][j]+tmp1);
	      // column - exchange:
	      if(tmp >= decomp_householder_eps)
		{
		  k=p[i];
		  p[i]=p[j];
		  p[j]=k;
		  for(k=0;k<n;k++)
		    {
		      tmp2 = A[k][i];
		      A[k][i]=A[k][j];
		      A[k][j]=tmp2;
		    } 
		  break;
		} // end of column exchange
	    } // end of column exchange if || A[.][j] || > eps
	  if( j==m ) // singular case
	    return i;
	} // singular case.
      break;
    case always:
      for(tmp2=0.0,i_max=i,k=i;k<m;k++)
	{
	  for(tmp1=0.0,j=i+1;j<n;j++)
	    tmp1+=sqr(A[k][j]);
	  tmp = sqrt(sqr(A[k][i])+tmp1);
	  if(tmp>tmp2)
	    { i_max=k; tmp2 = tmp; }
	}
      if(tmp2 < decomp_householder_eps)
	return i;
      // column exchange:
      if(i_max != i)
	{
	  j=p[i];
	  p[i] = p[i_max];
	  p[i_max]=j;
	  for(j=0;j<n;j++)
	    {
	      tmp2 = A[j][i];
	      A[j][i]=A[j][i_max];
	      A[j][i_max]=tmp2;
	    }
	}
      // computing norm of A[.][i]:
      for(j=i+1,tmp1=0.0;j<n;j++)
	tmp1+=A[j][i]*A[j][i];

      tmp = sqrt(A[i][i]*A[i][i]+tmp1);
      break;
    default:
      exit(-1); // This must never happen!
    }

      // Householder reflection:
      d[i] = ( A[i][i] > 0.0 ? -tmp : tmp );
      
      A[i][i] -= d[i];
      tmp = sqrt(2.0/(A[i][i]*A[i][i]+tmp1));

      // scaling... of the reflection vector h:
      for(j=i;j<n;j++)
	  A[j][i] *= tmp;

      // Householder reflection:
      for(j=i+1;j<m;j++)
	{
	  // scalar product h'*A[.][j]
	  for(tmp2=0.0,k=i;k<n;k++)
	    tmp2 += A[k][i] * A[k][j];

	  // reflection Anew[.][j] = A[.][j] - h*(h'*A[.][j])
	  for(k=i;k<n;k++)
	    A[k][j] -= A[k][i] * tmp2;
	} // Householder reflection
    } // main loop

  return m; // =m Matrix A has maximal rank.
}



int lin_solve_householder(int n,
			   int m,
			   double **A,
			   double *d,
			   int* p,
			   double* xy)
  /*
    n: input: number of rows of A,
    m: input: number of columns of A,
    A: input: (most of the :-) Householder decomposition QR of A
    d: input: diagonal of R
    p: input: permutation vector. If p[0] < 0 then rank(A)<m.
    xy: input: right hand side of the linear system A x = y
        output: least square solution of the system A x = y
   */
{
  int i,j,k;
  double tmp;
  int rank;
  
  // Computation of the Householder-decomposition A=QR
  rank = decomp_householder(n,m,A,d,p);
    /*
      calculation of: Q'y
    */
  for(i=0;i<rank;i++)
    {
      for(j=i,tmp=0.0;j<n;j++)
	tmp+=A[j][i]*xy[j];
      for(j=i;j<n;j++)
	xy[j] -= A[j][i]*tmp;
    } // calculation of Q'y
  
  if(rank==m) { // rank(A) = m ...
    // solving Rx = Q'y (upper triangle)
    for(i=m-1;i>=0;i--)
      {
	for(tmp=xy[i],j=i+1;j<m;j++)
	  tmp -= A[i][j]*xy[j];
	xy[i] = tmp/d[i];
      }
  } // rank(A) = m
  else { // rank(A) < m
    /* 
       compute the solution of  |x|->min
	with the restriction  Rx = Q'y
     */
    for(i=0;i<rank;i++) // computation of S=R*R', Note: d[i]=R[i][i]
      for(j=i;j<rank;j++)
	{
	  if(i==j) tmp = sqr(d[i]);
	  else tmp = A[i][j]*d[j];
	  for(k=j+1;k<m;k++) tmp+=A[i][k]*A[j][k];
	  A[j][i]=tmp; // store the lower triangle of S into A (overwriting Q)
	}
    // compute Lagrangian coefficiens c via Cholesky decomposition S = L*L'
    for(i=0;i<rank;i++)
      {
	for(j=0;j<i;j++)
	  {
	    for(tmp=A[i][j],k=0;k<j;k++)
	      tmp -= A[i][k]*A[j][k];
	    A[i][j]=tmp/A[j][j];
	  }
	for(tmp=A[i][i],k=0;k<i;k++)
	  tmp -= sqr(A[i][k]);
	A[i][i]=sqrt(tmp);
      }
    // solving L*(L'* c) = xy for (L'*c):
    for(i=0;i<rank;i++)
      {
	for(j=0;j<i;j++)
	  xy[i]-=A[i][j]*xy[j];
	xy[i] /= A[i][i];
      }
    // solving L'* c = xy for c:
    for(i=rank-1;i>=0;i--)
      {
	for(j=i+1;j<rank;j++)
	  xy[i]-=A[j][i]*xy[j];
	xy[i] /= A[i][i];
      }
    // Now the first (rank) components are the Lagrangian coefficients.
    // Computing x of Ax ~ y with ( |x| -> min ) via x = R' * c
    for(i=m-1;i>=rank;i--)
	for(xy[i]=0.0,j=0;j<rank;j++)
	  xy[i] += A[j][i]*xy[j];
    for(i=rank-1;i>=0;i--)
      {
	for(tmp=d[i]*xy[i],j=i-1;j>=0;j--)
	  tmp+=A[j][i]*xy[j];
	xy[i]=tmp;
      }
  } // rank(A) < m

  for(i=0;i<m;i++) d[i]=xy[i];
  for(i=0;i<m;i++) xy[p[i]]=d[i];

  return rank;
} // lin_solve_householder


t_interpolate::t_interpolate(int n)
  {
    my_n = n;
    my_w = new double [my_n]; // w stands for workspace.
    my_w2 = new double [my_n];
  }
  
t_interpolate::~t_interpolate()
  {
    delete [] my_w;
    delete [] my_w2;
  }

double t_interpolate::operator () (
		      int n,
		      double t,
		      double *t_points,
		      double *x_points
		      )
  { // n= Anzahl der Stuetzstellen
    int k,l;
    if( my_n < n )
      {
	delete [] my_w;
	delete [] my_w2;
	my_n = n;
	my_w = new double [my_n];
	my_w2 = new double [my_n];
      }
    
    for(k=0; k<n; k++)
      my_w[k] = x_points[k];
    
    for(l=1;l<n;l++)
      for(k=0;k<n-l;k++)
	my_w[k] = (my_w[k]*(t-t_points[k+l])
		      + my_w[k+1]*(t_points[k]-t))/
	  (t_points[k]-t_points[k+l]);
    
    return my_w[0];
  }

double t_interpolate::operator () (
		      int n,
		      double t,
		      double *t_points,
		      double *x_points,
		      double *dxdt
		      )
  { // n= Anzahl der Stuetzstellen
    int k,l;
    if( my_n < n )
      {
	delete [] my_w;
	delete [] my_w2;
	my_n = n;
	my_w = new double [my_n];
	my_w2 = new double [my_n];
      }
    
    for(k=0; k<n; k++)
      {
	my_w[k] = x_points[k];
	my_w2[k] = 0.0;
      }
    
    for(l=1;l<n;l++)
      for(k=0;k<n-l;k++)
	{
	  my_w2[k] = (
		      my_w2[k]*(t-t_points[k+l])
		     + my_w2[k+1]*(t_points[k]-t)
		      +my_w[k]-my_w[k+1]
		      )/(t_points[k]-t_points[k+l]);
	  
	  my_w[k] = (
		     my_w[k]*(t-t_points[k+l])
		     + my_w[k+1]*(t_points[k]-t)
		     )/(t_points[k]-t_points[k+l]);
	}

    *dxdt = my_w2[0];

    return my_w[0];
  }

t_interpolate interpolate;

