/*
  Explicit embedded Runge Kutta methods for solving ode's
  d/dt x(t) = f( t, x(t) );
  (with adaptive step size control)

  For an example try
  g++ -DMAIN -g tobiRungeKutta.cc -ltobiMath -o tobiRungeKutta
  tobiRungeKutta
*/
#include"tobiRungeKutta.hh"

t_runge_kutta::t_tab
t_runge_kutta ::
fehlberg(
	 2,
	 (double[]){1.0,0.5},
	 (double[]){1.0,
		      0.25,0.25},
	 (double[]){0.5,0.5},
	 (double[]){1.0/6.0,1.0/6.0,2.0/3.0}),
t_runge_kutta ::
sarafyan(
	 5,
	 (double[]){0.5,0.5,1.0,2.0/3.0,1.0/5.0},
	 (double[]){
	   0.5,
	     0.25,0.25,
	     0,-1.0,2.0,
	     7.0/27.0,10.0/27.0,0.0,1.0/27.0,
	     28.0/625.0,-125.0/625.0,546.0/625.0,54.0/625.0,-378.0/625.0
	     },
	 (double[]){1.0/6.0,0.0,2.0/3.0,1.0/6.0,0.0},
	 (double[]){14.0/336.0,0.0,0.0,35.0/336.0,162.0/336.0,125.0/336.0}
	 );

bool t_runge_kutta::step(double _t_max)
{
  int i,j;
  double *f_tab;
  double err;
  t_v tmp_v(n), x_order1(n), x_order2(n);
    
  if( fabs(stepsize) > fabs(t - _t_max) )
    stepsize = _t_max-t;
  
  while(1) { // decreasing stepsize while error larger err_max

    f_tab = tab.p_f;
    pf[0] = f(t,x);

    for(i=1;i<=tab.order;i++)
      {
	tmp_v = 0.0;
	for(j=0;j<i;j++)
	    tmp_v += f_tab[j] * pf[j];
	pf[i] = f(t+stepsize*tab.p_t[i-1], x + stepsize*tmp_v);
	f_tab += i;
      }
    
    tmp_v = 0.0;
    for(i=0;i<tab.order;i++)
      tmp_v += tab.p_dx_1[i] * pf[i];
    
    x_order1 = x+stepsize*tmp_v;

    tmp_v = 0.0;
    for(i=0;i<tab.order+1;i++)
      tmp_v += tab.p_dx_2[i] * pf[i];

    x_order2 = x+stepsize*tmp_v;

    // step size control:
    err = norm(x_order1-x_order2)/stepsize;

    if( err <= err_max ) break;

    stepsize *= pow(err_fac*err_max/err,1.0/tab.order);

    if(stepsize < step_min) return false;
  } // End of step size control.

  if( err < err_min_fac*err_max )
    {
      double tmp = stepsize;
      stepsize *= pow(err_fac*err_max/err,1.0/tab.order);
      if( stepsize < tmp )
	{
	  std::cerr << "\n t_runge_kutta: Outch!"<< std::endl;
	  abort();
	}
      if( stepsize > step_max ) stepsize = step_max;
    }

  t = t+stepsize;
  x = x_order2;
  return true;
} // end of t_runge_kutta::step()

#ifdef MAIN

#include<iostream>
#include<fstream>

struct t_hopf : public t_runge_kutta
{
  t_hopf() : t_runge_kutta(2, // number of variables
			   0.0, // start time
			   100, // end time
			   0.1, // initial step size
			   1.0 // maximal step size
			   )
  {
    x[0]=0.0; x[1]=10.0;
    err_max=1.0e-3;
    set(sarafyan);
  }
  
  t_v f(double t, const t_v& x)
  {
    t_v v(2);
    double amp=1.0-norm(x);

    v[0] = amp * x[0] - x[1];
    v[1] = x[0] + amp * x[1];

    return v;
  }
} s;

main()
{
  ofstream os("test.dat");
  ofstream os2("test2.dat");
  while( s.t < s.t_max ){
    s.step();
    os << s.t << ' ' << s.x[0] << ' ' << s.x[1] << ' ' << s.stepsize << endl;
  }

  // The rough method to reset: Just set time to zero. The initial
  // conditions are taken from the former integration.
  s.t = 0.0;

  while( s.t < s.t_max ){
    s.step();
    os2 << s.t << ' ' << s.x[0] << ' ' << s.x[1] << ' ' << s.stepsize << endl;
  }
}
#endif


